import ldap,sys,ldif,ldap.modlist
import syslog
import binascii
import unicodedata

# LOGLEVEL can be set from 0 to 100. Higher is the value, more verbose will be the script
LOGLEVEL=0

# The prefix string in the syslogs
LOGIDENT="ad_import"

# If set to 1, output ldif to stdout
GENLDIF=0

# if set to 1, directly synchronize the two LDAP directories
SYNCHRONIZE=1

# if set to 1, output messages to stdout
LOGSTDOUT=0

# if set to 1, force a full synchronization
FULLSYNC=0

# some default values, in case of the attributes are missing in the source directory
computerUidNumberBase=5000
computerGidNumber="5000"
dummyUidNumber=40000
dummyGidNumber=40000
smbDomainName="mydomain"
smbDomainSID=""
defaultPassword="xQYL2KP333duo"
maxUSNChanged=0
lastMaxUSNChanged=0

# global constants, don't modify please
LOG_ALWAYS=-1
dicttype=type({})
arraytype=type([])
fncttype=type(lambda x: x)
strtype=type("")
inttype=type(0)
nonetype=type(None)
ldif_writer=ldif.LDIFWriter(sys.stdout)


# log function
# It will log the message in the syslog utility, depending on the Level and the LOGLEVEL value
# it will also log message to stdout if GENLDIF or LOGSTDOUT is set
def LOG(Prio, Level, Message):
    if LOGLEVEL<=Level:
       return 1
    syslog.syslog(Prio, Message)   
    if GENLDIF==1 or LOGSTDOUT==1:
       print "###",Message
    return 1

# extract a substring 
def substr(aString, start, n):
    return aString[start:start+n]

# remove accents and other strange characters
def normalizeString(s):
    if isinstance(s,str):
       s = unicode(s,"utf8","replace")
    s=unicodedata.normalize('NFD',s)
    return s.encode('ascii','ignore')


# transform an AD binary SID to a text SID
def binSid2TxtSid(binSid):
    ''' according to http://forum.java.sun.com/thread.jspa?threadID=585031&tstart=0 '''
    ''' and http://udf.isorg.com/UDFs/83254.html (the first link is wrong for the count value'''
### objectSid: version|sub-authority count|identifier-authority 48bis|{sub authorities|xcount}
###            version : 1st byte
###            count 2nd byte
###            id : bytes 3-8 
###            sub auth 1: 9-12 
###            sub auth 2: 13-16 
###            sub auth n: (8+(n-1)*4)-(12+(n-1)*4)
###            d36b4a5c
###            b0050000 
### SambaSID : S-01-0-21-930152448-2096052081-201686255-3008433157
### objectSid: 01020000000000052000000027020000
### SambaSID : S-1-5-32-551
### objectSid: 0105000000000005150000007137ef7c050c51b3d36b4a5c90040000
### SambaSID : S-1-5-21-2096052081-3008433157-1548381139-1168

    byteSid=binascii.b2a_hex(binSid)
    LOG(syslog.LOG_DEBUG, 8, "objectSid: "+byteSid)
    strSID='S-'+str(int(substr(byteSid,0,2), 16))
    rid=substr(byteSid,4,12)
    LOG(syslog.LOG_DEBUG, 8, "      rid: '"+rid+"'")
    strSID=strSID+'-'+str(int(rid,16))
    count=int(substr(byteSid,2,2),16)
    LOG(syslog.LOG_DEBUG, 8, "    count: '"+str(count)+"'")
    for i in range(count):
       rid=substr(byteSid, 22+i*8, 2)+substr(byteSid, 20+i*8, 2)+substr(byteSid, 18+i*8, 2)+substr(byteSid, 16+i*8, 2)
       LOG(syslog.LOG_DEBUG, 8, "      rid: '"+rid+"'")
       strSID=strSID+'-'+str(int(rid, 16))
    LOG(syslog.LOG_DEBUG, 8, "SambaSID : "+strSID)
    return strSID

def getRid(binSid):
    byteSid=binascii.b2a_hex(binSid)
    last=int(substr(byteSid,2,2),16)-1
    rid=substr(byteSid, 22+last*8, 2)+substr(byteSid, 20+last*8, 2)+substr(byteSid, 18+last*8, 2)+substr(byteSid, 16+last*8, 2)
    return int(rid,16)
    
# try to translate AD groupType to sambaGroupType
def getSambaGroupType(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    '''
    SID_NAME_USE_NONE = 0,/* NOTUSED */
    SID_NAME_USER    = 1, /* user */
    SID_NAME_DOM_GRP = 2, /* domain group */
    SID_NAME_DOMAIN  = 3, /* domain: don't know what this is */
    SID_NAME_ALIAS   = 4, /* local group */
    SID_NAME_WKN_GRP = 5, /* well-known group */
    SID_NAME_DELETED = 6, /* deleted account: needed for c2 rating */
    SID_NAME_INVALID = 7, /* invalid account */
    SID_NAME_UNKNOWN = 8  /* oops. */
    '''
    if val=="-2147483643":
       ret="5"
    else:
       ret="2"
  
    if ret=="2":
       if getRid(entry[1]["objectSid"][0])==515:
          global computerGidNumber
          computerGidNumber=entry[1]["msSFU30GidNumber"][0]

    return ret

# translate an objectSID to a sambaSID
def getSambaSID(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    return binSid2TxtSid(val)

# get the default samba domain SID and add the primaryGroupID we got from AD
def getSambaPrimaryGroupSID(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    return smbDomainSID+'-'+str(val)

# get the default samba domain name
def getSambaDomainName(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    return smbDomainName

# set the default samba domain nae and SID from the AD domain
def initSambaDomainName(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    global smbDomainName
    global smbDomainSID
    smbDomainName=entry[1]["dc"][0].upper()
    smbDomainSID=mapped["sambaSID"][0]
    return smbDomainName

# set the sambaDomain entry DN
def setSambaDomainNameDN(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    smbDomainName=initSambaDomainName(l, srcbase, entry, attr, val, mapped, dstbase, dstldap)
    return "sambaDomainName="+smbDomainName+","+srcbase


# Samba Account Flags
    '''
       D: Account is disabled
       H: home Directory is required
       I: inter domain trust account
       L: account has been auto locked
       M: Microsoft Network Service Logon Account
       N: password NOT required
       S: server trust account
       T: temporary duplicate account entry
       U: normal User account
       W: workstation trust account
       X: password does not expire
    '''
# userAccountControl attribute:
# http://support.microsoft.com/kb/305144


def userAccountControl2SambaAcctFlags(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    ret=''
    Flags={2: 'D', 32: 'N', 8: 'H', 2048: 'I', 131072: 'M', 4096: 'W', 256: 'T', 
           8192: 'S', 512: 'U', 65536: 'X', 16: 'L'}
    ival=int(val)
    for f in Flags:
       if ival&f:
          ret=ret+Flags[f]
    
    return '['+ret+substr('           ',0,11-len(ret))+']'

def normalizeValue(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    return normalizeString(val)

def readMaxUSNChanged(dstldap, dstbase):
    global maxUSNChanged
    global lastMaxUSNChanged
    weLookFor="uSNChanged"
    try:
       r2=dstldap.search_s(dstbase, ldap.SCOPE_BASE) 
    except ldap.NO_SUCH_OBJECT, error_message:
      LOG(syslog.LOG_WARNING, LOG_ALWAYS, "readMaxUSNChanged: "+str(error_message))
      return 0
    except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "readMaxUSNChanged: "+str(error_message))
      return 0

    if(len(r2)>0 and isinstance(r2[0][1], dict) and r2[0][1].has_key(weLookFor)):
      maxUSNChanged=int(r2[0][1][weLookFor][0])
      lastMaxUSNChanged=maxUSNChanged
      LOG(syslog.LOG_DEBUG, 3, "read maxUSNChanged="+str(maxUSNChanged))
    return

def writeMaxUSNChanged(dstldap, dstbase):
    global maxUSNChanged
    weLookFor="uSNChanged"
    try:
       r2=dstldap.search_s(dstbase, ldap.SCOPE_BASE) 
    except ldap.NO_SUCH_OBJECT, error_message:
      LOG(syslog.LOG_WARNING, LOG_ALWAYS, "writeMaxUSNChanged: "+str(error_message))
      return 0
    except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "writeMaxUSNChanged: "+str(error_message))
      return 0

    LOG(syslog.LOG_DEBUG, 3, "write maxUSNChanged="+str(maxUSNChanged))

    try:
      dstldap.modify_s(dstbase, [(ldap.MOD_REPLACE, "uSNChanged", str(maxUSNChanged))])
    except ldap.NO_SUCH_OBJECT, error_message:
      LOG(syslog.LOG_WARNING, LOG_ALWAYS, "writeMaxUSNChanged: "+str(error_message))
    except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "writeMaxUSNChanged: "+str(error_message))
    return 0

def setMaxUSNChanged(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    global maxUSNChanged
    v=int(val)
    if v>maxUSNChanged:
       maxUSNChanged=v
    return val

def genUid(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    LOG(syslog.LOG_WARNING, LOG_ALWAYS, 'uid set to "bidon" for '+entry[0])
    return "bidon"

def getPoolUidNumber(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    LOG(syslog.LOG_WARNING, LOG_ALWAYS, 'uid set to "bidon" for '+entry[0])
    return "10000"

def getPoolGidNumber(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    LOG(syslog.LOG_WARNING, LOG_ALWAYS, 'uid set to "bidon" for '+entry[0])
    return "10000"

def genGidNumber(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    global dummyGidNumber
    ret=str(dummyGidNumber)
    LOG(syslog.LOG_WARNING, LOG_ALWAYS, "gidNumber set to "+ret+" for "+entry[0])
    dummyGidNumber=dummyGidNumber+1
    return ret

def genHome(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    # LOG(syslog.LOG_WARNING, LOG_ALWAYS, "home set to /tmp for "+entry[0])
    return "/tmp"

def genLoginShell(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    #LOG(syslog.LOG_WARNING, LOG_ALWAYS, "loginShell set to /bin/sh for "+entry[0])
    return "/bin/bash"

def genUidNumber(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    global dummyUidNumber
    ret=str(dummyUidNumber)
    LOG(syslog.LOG_WARNING, LOG_ALWAYS, "uid set to "+ret+" for "+entry[0])
    dummyUidNumber=dummyUidNumber+1
    return ret

def getMemberUid(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    weLookFor="msSFU30Name"
    try:
       r2=l.search_s(val, ldap.SCOPE_BASE) 
    except ldap.NO_SUCH_OBJECT, error_message:
      LOG(syslog.LOG_WARNING, LOG_ALWAYS, "getMemberUid: "+str(error_message))
      return 0 
    except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "getMemberUid: "+str(error_message))
      return 0 

    LOG(syslog.LOG_DEBUG, 10, "r2="+str(r2)) 

    if(len(r2)>0 and isinstance(r2[0][1], dict) and r2[0][1].has_key(weLookFor)):
       return normalizeString(r2[0][1][weLookFor][0])

    return 0

def getMember(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    ret=[]
    if not entry[1].has_key("member"):
       return ret
    for i in entry[1]["member"]:
       ret=ret+[mapdn(i, srcbase, dstbase)]
    return ret
    
def getSambaSIDList(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    weLookFor="objectSid"
    ret=[]
    if not entry[1].has_key("member"):
       return ret
    for i in entry[1]["member"]:
       try:
          r2=l.search_s(i, ldap.SCOPE_BASE) 
       except ldap.NO_SUCH_OBJECT, error_message:
         LOG(syslog.LOG_WARNING, LOG_ALWAYS, "getSambaSIDList: "+str(error_message))
         pass
       except ldap.LDAPError, error_message:
         LOG(syslog.LOG_ERR, LOG_ALWAYS, "getMemberUid: "+str(error_message))
         pass 

       LOG(syslog.LOG_DEBUG, 10, "r2="+str(r2)) 
       if(len(r2)>0 and isinstance(r2[0][1], dict) and r2[0][1].has_key(weLookFor)):
         ret=ret+[binSid2TxtSid(r2[0][1][weLookFor][0])]

    return ret
    
def getDstVal(l, srcbase, entry, weLookFor, val, mapped, dstbase, dstldap):
    try:
       r2=dstldap.search_s(mapdn(entry[0], srcbase, dstbase), ldap.SCOPE_BASE) 
    except ldap.NO_SUCH_OBJECT, error_message:
      LOG(syslog.LOG_WARNING, LOG_ALWAYS, "getDstVal: "+str(error_message))
      return 0 
    except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "getDstVal: "+str(error_message))
      return 0 

    LOG(syslog.LOG_DEBUG, 10, "r2="+str(r2)) 

    if(len(r2)>0 and isinstance(r2[0][1], dict) and r2[0][1].has_key(weLookFor)):
       return r2[0][1][weLookFor][0]

    return 0

def getPassword(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    if val=="ABCD!efgh12345$67890" or val=="":
       ret=defaultPassword
    else:
       ret=val
    if entry[1].has_key("userAccountControl"):
       userAccountControl=int(entry[1]["userAccountControl"][0])
       disabled=userAccountControl&2
    else:
       disabled=0
    if disabled:
       disabled='!'
    else:
       disabled=''
    return "{CRYPT}"+disabled+ret 

def getMailNickName(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    if entry[1].has_key("mailNickname"):
       return entry[1]["mailNickname"]
    else:
       if mapped.has_key("uid"):
          return mapped["uid"]
       else:
          return "void"

def getComputerName(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    if entry[1].has_key("name"):
       return entry[1]["name"]
    else:
       if mapped.has_key("cn"):
          return mapped["cn"]
       else:
          return "void"

def getComputerUid(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    if entry[1].has_key("cn"):
       return entry[1]["cn"][0]+'$'
    else:
       return "void$"

def getComputerUidNumber(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    return str(computerUidNumberBase+getRid(entry[1]["objectSid"][0]))

def getComputerGidNumber(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    return computerGidNumber

def getL(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    mapL={"Paris": "Paris", "Londres": "London", "Hong-Kong": "Honk-Kong", "New-York": "New-York"}
    for i in mapL:
      if entry[0].find(i)>-1:
         return  mapL[i]
    return "Unknown"

def normalizeDN(dn):
    for i in ["CN","DN","OU","SN","DC"]:
       il=i.lower()+'='
       dn=dn.replace(i+'=',il)
    return dn

def mapdn(dn, src, dst):
    dn=normalizeDN(dn)
    src=normalizeDN(src)
    dst=normalizeDN(dst)
       
    if src==dst:
       ret=dn
    else:
       ret=dn.replace(src, dst)
    LOG(syslog.LOG_DEBUG, 10, "mapdn: src="+src+", dst"+dst)
    LOG(syslog.LOG_DEBUG, 10, "mapdn: "+dn+"-->"+ret)
    return ret
    
def getDN(l, srcbase, entry, attr, val, mapped, dstbase, dstldap):
    return mapdn(entry[0], srcbase, dstbase )

def Synchronize(dstldap, dn, srcentry):
    if len(srcentry)==0:
       return ""

    LOG(syslog.LOG_DEBUG, 5, "Synchronize: "+dn)

    try:
       r=dstldap.search_s(dn, ldap.SCOPE_BASE)
    except ldap.NO_SUCH_OBJECT, error_message:
      r=[]
    except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "Synchronize: search_s "+str(error_message))
      r=[]

    if(len(r)==0):
      ml=ldap.modlist.addModlist(srcentry)
      LOG(syslog.LOG_DEBUG, 5, "Synchronize: addModList="+str(ml))
      try:
         dstldap.add_s(dn, ml)
      except ldap.LDAPError, error_message:
         LOG(syslog.LOG_ERR, LOG_ALWAYS, "Synchronize: add_s, dn="+dn+" "+str(error_message))
    else:
      if(type(r[0][0])==strtype and isinstance(r[0][1], dict)):
        LOG(syslog.LOG_DEBUG, 5, "Synchronize: src="+str(srcentry))
        LOG(syslog.LOG_DEBUG, 5, "Synchronize: dst="+str(r[0][1]))

        ml=ldap.modlist.modifyModlist(r[0][1], srcentry)

        LOG(syslog.LOG_DEBUG, 5, "Synchronize: modifyModList="+str(ml))
        try:
	  dstldap.modify_s(dn, ml)
        except ldap.LDAPError, error_message:
          LOG(syslog.LOG_ERR, LOG_ALWAYS, "Synchronize: modify_s, dn="+dn+" "+str(error_message))
    return 1
  
def ImportError():
    LOG(syslog.LOG_ERR, LOG_ALWAYS, "dn="+dn)
    LOG(syslog.LOG_ERR, LOG_ALWAYS, "attr="+attr)
    LOG(syslog.LOG_ERR, LOG_ALWAYS, "currentvalue="+currentvalue)
    LOG(syslog.LOG_ERR, LOG_ALWAYS, "mappage non traite")
    return 0

def Import(l, srcbase, searchbase, scope, dstldap, dstbase, 
           filter, attrs, attrmap, valuemap, mandatory):
    """ l is the ldap object for the source server 
        srcbase is the basedn of the hierarchy in the source ldap
        searchbase is the basedn of the objects we want to import 
        dstldap is the ldap object for the destination server
        attrs is an array of string : they are the attributes we want to get from the source server
        attrmap is a dictionary object:
           the keys are the name of one of the attributes we requested,
           the data associated is a string representing the name of this attribute in the destination server
           so, if you have {"foo": "bar"}, it means the attribute "foo" will be mapped to "bar"
        valuemap is a dictionary object:
           the keys are the name of the destination attribute (so, after the mapping done with the dict above)
           the values are either
               - a function
                 in this case, the initial value is replaced with the return value of the function
	         the function can return either a string or an array
               - a dictionary
	         the keys are the possible values we look for this attribute
		 the values are either a string or an array 
               - a string
        mandatory is a dictionary object:
           the keys are the name of the attribute we absolutly need in the destination server
           the values are either a string, an array or a function which can return 
           what we want for these attributes
    """

    try:
       result_set=l.search_s(searchbase, scope, filter, attrs)
    except ldap.NO_SUCH_OBJECT, error_message:
       LOG(syslog.LOG_WARNING, LOG_ALWAYS, "Import: search_s "+str(error_message))
       return 1 
    except ldap.LDAPError, error_message:
       result_set=[]
       LOG(syslog.LOG_ERR, LOG_ALWAYS, "Import: search_s "+str(error_message))

    if len(result_set)==0:
       return 1

    for i in range(len(result_set)):
       if(type(result_set[i][0])==strtype and isinstance(result_set[i][1], dict)):
         export={}
         dn=result_set[i][0]

	 LOG(syslog.LOG_DEBUG, 8, "Import: "+str(result_set[i][1]))

         for k in result_set[i][1].keys():
           if(attrmap.has_key(k)):
             attr=attrmap[k]
           else:
             attr=k

	   LOG(syslog.LOG_DEBUG, 5, "Import: attribute "+k+" mapped to "+attr)

           if(valuemap.has_key(attr)):
             vm=valuemap[attr]
           else:
             vm=0  

           export[attr]=[]
           for n in range(len(result_set[i][1][k])): 
	      currentvalue=result_set[i][1][k][n]
              if(vm==0):
                value=currentvalue
              else:
                if(type(vm)==fncttype):
                  value=vm(l, srcbase, result_set[i], attr, currentvalue, export, dstbase, dstldap)
                else:
                  if(type(vm)==dicttype):
                    if(vm.has_key(currentvalue)):
		      intvalue=vm[currentvalue]
                      if(type(intvalue)==strtype):
                        value=intvalue
                      else:
                        if(type(intvalue)==arraytype):
		          value=intvalue
                        else:
                          if type(intvalue)==nonetype:
                             value=None
                          else:
                             ImportError(dn, attr, currentvalue)
                    else:
                      value=currentvalue
                  else:
                    ImportError(dn, attr, currentvalue)

              if(not k=="objectSid"):
                LOG(syslog.LOG_DEBUG, 5, "Import: currentvalue="+currentvalue)
 	        LOG(syslog.LOG_DEBUG, 5, "Import: ------------>"+str(value))

              if type(value)!=nonetype:
                 if(type(value)==arraytype):
	           export[attr]=export[attr]+value
                 else:
                   if(type(value)==strtype):
	             export[attr]=export[attr]+[value]

	      LOG(syslog.LOG_DEBUG, 12, "Import: export="+str(export))

         for m in mandatory:
            if not export.has_key(m):
              if(type(mandatory[m])==strtype):
                value=mandatory[m]
              else:
                if(type(mandatory[m])==fncttype):
                  value=mandatory[m](l, srcbase, result_set[i], m, "", export, dstbase, dstldap)
                else:
                  value=[]
              if m=="dn":
                 if type(value)==strtype:
                    dn=value                
              else:
                 if type(value)==strtype or type(value)==arraytype:
                    if type(value)==strtype:
                       value=[value]
                    export[m]=value

         LOG(syslog.LOG_DEBUG, 12, "Import: export="+str(export))

         if GENLDIF==1:
            ldif_writer.unparse(mapdn(dn, srcbase, dstbase), export)

         if SYNCHRONIZE==1:
            Synchronize(dstldap, mapdn(dn, srcbase, dstbase), export) 

    return 1

def ImportBase(dstldap, s, d, samba):
    levels=ldap.explode_dn(d,1)
    nblevel=len(levels)
    for n in range(nblevel):
       dn=""
       for sb in range(n+1):
          dn="dc="+levels[nblevel-1-sb]+dn
          if not sb==n:
             dn=","+dn

       entry={"objectClass": ["top", "domain", "syncAD"], "dc": [levels[nblevel-1-n]],
                               "uSNCreated": ["0"], "uSNChanged": [str(maxUSNChanged)]}
       if SYNCHRONIZE:
	  Synchronize(dstldap, dn, entry)

       if GENLDIF==1:
          ldif_writer.unparse(dn, entry)

    levels=ldap.explode_dn(samba, 1)
    entry={"objectClass": ["top", "container"], "cn": [levels[0]]}
    if SYNCHRONIZE:
       Synchronize(dstldap, samba, entry)

    if GENLDIF==1:
       ldif_writer.unparse(samba, entry)

    return ""

def CheckRemoved(dstldap, srcldap, srcbase, dstbase, excepted):
   # On recupere juste la liste des DNs
   try:
      r=dstldap.search_s(dstbase, ldap.SCOPE_SUBTREE, "(&(objectClass=syncAD)(!(uSNCreated=0)))", ["NOATTR"])
   except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "CheckRemoved:"+str(error_message))
      return 1

   for i in range(len(excepted)):
      excepted[i]=normalizeDN(excepted[i])

   for i in r:
      found=0
      for e in excepted:
         if normalizeDN(i[0]).find(e)>=0:
	    found=1
	    break

      if not found:
         try:
           r=srcldap.search_s(mapdn(i[0], dstbase, srcbase), ldap.SCOPE_BASE, "(objectClass=*)", ["NOATTR"])
         except ldap.NO_SUCH_OBJECT, error_message:
           LOG(syslog.LOG_INFO, LOG_ALWAYS, "deleting "+i[0])
           try:
             dstldap.delete(i[0])
           except ldap.LDAPError, error_message:
             LOG(syslog.LOG_ERR, LOG_ALWAYS, "CheckRemoved: could not delete "+str(error_message))
         except ldap.LDAPError, error_message:
           LOG(syslog.LOG_ERR, LOG_ALWAYS, "CheckRemoved: search_s "+str(error_message))
           r=[]

   return 1

def makeFilter(aFilter):
    aFilter="(&"+aFilter+"(uSNChanged>="+str(lastMaxUSNChanged)+"))"
    LOG(syslog.LOG_DEBUG, 4, "aFilter="+aFilter)
    return aFilter

def main_1(): 
   srcuri="ldap://192.168.0.30"
   srcadmin="cn=openldap,OU=Gestion Admin,OU=Utilisateurs,dc=mydomain,dc=local"
   srcpassword="openldap"
   srcbasedn="dc=mydomain,dc=local"

   dsturi="ldap://127.0.0.1"
   dstadmin="cn=admin,dc=openldap,dc=mydomain,dc=local"
   dstpassword="MyPassword"
   dstpassword="soleil"
   dstbasedn="dc=openldap,dc=mydomain,dc=local"
   dstsamba="cn=sambaDomains,"+dstbasedn

   groupbase=srcbasedn
   oubase=srcbasedn
   userbase=srcbasedn

   syslog.openlog("import_ad")

   # To avoid the DSID-0C090627 error from AD.
   ldap.set_option(ldap.OPT_REFERRALS, 0)

   try:
      srcldap=ldap.initialize(srcuri)
      srcldap.simple_bind_s(srcadmin, srcpassword)
   except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "main: cannot connect and/or bind to source server,"+str(error_message))
      return 1

   try:
      dstldap=ldap.initialize(dsturi)
      dstldap.simple_bind_s(dstadmin, dstpassword) 
   except ldap.LDAPError, error_message:
      LOG(syslog.LOG_ERR, LOG_ALWAYS, "main: cannot connect and/or bind to destination server,"+str(error_message))
      return 1
 
   if not FULLSYNC:
      readMaxUSNChanged(dstldap, dstbasedn)

   CheckRemoved(dstldap, srcldap, srcbasedn, dstbasedn, [dstsamba])

   ImportBase(dstldap, srcbasedn, dstbasedn, dstsamba)

   Import(srcldap, srcbasedn, srcbasedn, ldap.SCOPE_SUBTREE, dstldap, dstsamba,
          "(objectClass=domainDNS)",
	  ["objectClass", "objectSid", "dc", "maxPwdAge", "minPwdAge", "minPwdLength", "pwdHistoryLength",
	   "uSNChanged", "uSNCreated"], 
           {"dc": "sambaDomainName", "objectSid": "sambaSID", "nextRid": "sambaNextRid",
            "minPwdLength": "sambaMinPwdLength", "pwdHistoryLength": "sambaPwdHistoryLength", 
            "maxPwdAge": "sambaMaxPwdAge", "minPwdAge": "sambaMinPwdAge" }, 
          {"objectClass": {"domain": "syncAD", "domainDNS": ["sambaDomain", "sambaUnixIdPool"]}, 
           "sambaSID": getSambaSID, "uSNChanged": setMaxUSNChanged}, 
          {"dn": setSambaDomainNameDN, "uidNumber": getPoolUidNumber, "gidNumber": getPoolGidNumber})
	     
   Import(srcldap, srcbasedn, oubase, ldap.SCOPE_ONELEVEL, dstldap, dstbasedn,
          makeFilter("(|(objectClass=container)(objectClass=builtinDomain))"),
	  ["objectClass", "cn", "uSNCreated", "uSNChanged"], {}, 
          {"objectClass": {"builtinDomain": "container", "top": ["top", "syncAD"]}, 
           "uSNChanged": setMaxUSNChanged}, {})
	     
   Import(srcldap, srcbasedn, oubase, ldap.SCOPE_SUBTREE, dstldap, dstbasedn,
	  makeFilter("(objectClass=organizationalUnit)"),
	  ["objectClass", "ou", "uSNChanged", "uSNCreated"], {},
          {"objectClass": {"top": ["top", "syncAD"]}, 
           "uSNChanged": setMaxUSNChanged}, {})
	     
   Import(srcldap, srcbasedn, groupbase, ldap.SCOPE_SUBTREE, dstldap, dstbasedn,
	  makeFilter("(&(objectClass=group)(msSFU30GidNumber=*))"),
          ["objectClass", "cn", "msSFU30GidNumber", "member", "msSFU30PosixMember",
           "objectSid", "groupType", "displayName", "description", "uSNChanged", "uSNCreated"],
          {"msSFU30GidNumber": "gidNumber", "member": "memberUid", "msSFU30PosixMember": "memberUid",
           "objectSid": "sambaSID", "groupType": "sambaGroupType"}, 
          {"objectClass": {"group": ["posixGroup", "sambaGroupMapping", "sambaIdmapEntry",
                                     "confluenceGroup", "syncAD"]}, 
           "memberUid": getMemberUid, "uSNChanged": setMaxUSNChanged,
           "sambaSID": getSambaSID, "sambaGroupType": getSambaGroupType},
	  {"gidNumber": genGidNumber,
	   "member": getMember, "sambaSIDList": getSambaSIDList})
      
   for cb in ['ou=Ordinateurs,'+srcbasedn, 'ou=Computers,'+srcbasedn]:
      Import(srcldap, srcbasedn, cb, ldap.SCOPE_SUBTREE, dstldap, dstbasedn,
             makeFilter("(objectClass=computer)"),
             ["objectClass", "cn",  "name", "uSNChanged", "uSNCreated",
              "userPrincipalName", "objectSid", "l", "primaryGroupID", 
	      "displayName", "description", "userAccountControl" ],
             {"objectSid": "sambaSID", "homeDirectory": "sambaHomePath",
               "badPwdCount": "sambaBadPasswordCount", "userAccountControl": "sambaAcctFlags",
              "primaryGroupID": "sambaPrimaryGroupSID", "name": "sn" },
             {"objectClass": {"user":  ["posixAccount", "sambaSamAccount", "user", "syncAD", "sambaIdmapEntry", "shadowAccount"]},
              "userPassword": getPassword, "sambaSID": getSambaSID,
              "sambaPrimaryGroupSID": getSambaPrimaryGroupSID, "uSNChanged": setMaxUSNChanged,
	      "loginShell": genLoginShell, "sambaAcctFlags": userAccountControl2SambaAcctFlags}, 
             {"uidNumber": getComputerUidNumber, "uid": getComputerUid, "homeDirectory": genHome,
              "sn": getComputerName, "gidNumber": getComputerGidNumber, "l": getL, 
	      "sambaDomainName": getSambaDomainName} )

   # http://support.microsoft.com/kb/305144 to fill sambaAccountFlag and maybe some shadowAccount stuffs
   # userAccountControl contains things suchs as account locked, user or workstation account and so on
   for ub in ['ou=BFI,'+srcbasedn, 'ou=Utilisateurs,'+srcbasedn, 'cn=Users,'+srcbasedn]:
      Import(srcldap, srcbasedn, ub, ldap.SCOPE_SUBTREE, dstldap, dstbasedn,
             makeFilter("(&(objectClass=user)(msSFU30UidNumber=*))"),
             ["objectClass", "sAMAccountName", "cn",  "sn", "msSFU30GidNumber",
              "msSFU30Name", "msSFUNisDomain", "msSFU30Password", "msSFU30UidNumber",
              "msSFU30HomeDirectory", "msSFU30LoginShell", "userPrincipalName",
              "mailNickName", "proxyAddresses", "objectSid", "l", "givenName", "mail",
              "profilePath", "homeDirectory", "scriptPath", "homeDrive", "badPwdCount",
	      "badPasswordTime", "pwdLastSet", "primaryGroupID", "uSNChanged", "uSNCreated",
	      "displayName", "description", "userAccountControl" ],
             {"msSFU30GidNumber": "gidNumber", "msSFU30Password": "userPassword", 
              "msSFU30HomeDirectory": "homeDirectory", "msSFU30UidNumber": "uidNumber",
              "msSFU30LoginShell": "loginShell", "msSFU30Name": "uid",
	      "objectSid": "sambaSID", "homeDirectory": "sambaHomePath",
              "profilePath": "sambaProfilePath", "scriptPath": "sambaLogonScript",
              "homeDrive": "sambaHomeDrive", "badPwdCount": "sambaBadPasswordCount",
              "badPasswordTime": "sambaBadPasswordTime", "pwdLastSet": "sambaPwdLastSet",
              "primaryGroupID": "sambaPrimaryGroupSID", "userAccountControl": "sambaAcctFlags" },
             {"objectClass": {"user":  ["posixAccount", "sambaSamAccount", "shadowAccount",
                                        "msExchUser", "mailRecipient", "user", "User",
                                        "syncAD", "sambaIdmapEntry"]},
              "userPassword": getPassword, "sambaSID": getSambaSID,
              "sambaPrimaryGroupSID": getSambaPrimaryGroupSID, "sambaAcctFlags": userAccountControl2SambaAcctFlags,
	      "loginShell": genLoginShell, "uid": normalizeValue, 
              "homeDirectory": normalizeValue, "uSNChanged": setMaxUSNChanged}, 
             {"uidNumber": genUidNumber, "uid": genUid, "homeDirectory": genHome,
              "mailNickname": getMailNickName, "sAMAccountName": getMailNickName,
              "sn": getMailNickName, "userPassword": getPassword,
              "gidNumber": "10038", "l": getL, 
	      "sambaDomainName": getSambaDomainName, "distinguishedName": getDN,
              "sambaNTPassword": getDstVal, "sambaLMPassword": getDstVal,
              "sambaPasswordHistory": getDstVal} )

   writeMaxUSNChanged(dstldap, dstbasedn)
   syslog.closelog()


if __name__=='__main__':
    main_1()
